"""
This tests the files
  lib/task/eventcounter.py
"""
from .base import MyTestCase
from .test_lib_counter import increase_and_read
from privacyidea.lib.counter import increase, read
from privacyidea.lib.monitoringstats import get_values

from privacyidea.lib.task.eventcounter import EventCounterTask
from flask import current_app


class TaskEventCounterTestCase(MyTestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_00_read_eventcounter_to_monitoringstats(self):
        r = increase_and_read("counter1")
        self.assertEqual(r, 1)

        r = increase_and_read("counter2")
        self.assertEqual(r, 1)

        r = increase_and_read("counter1")
        self.assertEqual(r, 2)

        r = read("counter1")
        self.assertEqual(r, 2)

        et = EventCounterTask(current_app.config)
        params = {"event_counter": "counter1",
                  "stats_key": "C1",
                  "reset_event_counter": "True"}

        # Now we execute the task
        et.do(params)

        # The counter "counter1" should be reset
        self.assertEqual(read("counter1"), 0)

        # The value "2" should be written to the statistics key C1.
        stats = get_values("C1")
        self.assertTrue(len(stats), 1)
        self.assertTrue(stats[0][1], 2)

        # Now we increase the event counter again
        increase("counter1")

        # ..and run the event counter task
        et.do(params)
        stats = get_values("C1")
        self.assertTrue(len(stats), 2)
        self.assertTrue(stats[0][1], 2)
        self.assertTrue(stats[0][1], 1)



"""Add table 'radiusserver' to configure the settings of remote RADIUS servers.

Revision ID: 449903fb6e35
Revises: 4023571658f8
Create Date: 2016-02-19 15:41:22.196074

"""

# revision identifiers, used by Alembic.
revision = '449903fb6e35'
down_revision = '4023571658f8'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.create_table('radiusserver',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('identifier', sa.Unicode(length=255), nullable=False),
        sa.Column('server', sa.Unicode(length=255), nullable=False),
        sa.Column('port', sa.Integer(), nullable=True),
        sa.Column('secret', sa.Unicode(length=255), nullable=True),
        sa.Column('description', sa.Unicode(length=2000), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('identifier')
        )
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Table 'radiusserver' already exists.")
        else:
            print(exx)
    except Exception as exx:
        print ("Could not add table 'radiusserver'")
        print (exx)


def downgrade():
    op.drop_table('radiusserver')




#  2019-07-02 Friedrich Weber <friedrich.weber@netknights.it>
#             Add a central module for comparing two values
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
This module implements comparisons between two values.
It is tested in test_lib_utils_compare.py

In order to add a new comparator:
 1) add the comparator to COMPARATORS
 2) implement a comparison function and add it to COMPARATOR_FUNCTIONS
 3) add a description of the comparator to COMPARATOR_DESCRIPTIONS
"""
import csv
import logging
import re
from functools import wraps

from privacyidea.lib.framework import _

log = logging.getLogger(__name__)


class CompareError(Exception):
    """
    Signals that an error occurred when carrying out a comparison.
    The error message is not presented to the user, but written to the logfile.
    """
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return "CompareError({!r})".format(self.message)


def parse_comma_separated_string(input_string):
    """
    Parse a string that contains a list of comma-separated values and return the list of values.
    Each value may be quoted with a doublequote, and doublequotes may be escaped with a backslash.
    Whitespace immediately following a delimiter is skipped.
    Raise a CompareError if the input is malformed.
    :param input_string: an input string
    :return: a list of strings
    """
    # We use Python's csv module because it supports quoted values
    try:
        reader = csv.reader([input_string], strict=True, skipinitialspace=True, doublequote=False, escapechar="\\")
        rows = list(reader)
    except csv.Error as exx:
        raise CompareError("Malformed comma-separated value: {!r}".format(input_string, exx))
    return rows[0]


def _compare_equality(left, comparator, right):
    """
    Return True if two values are exactly equal, according to Python semantics.
    """
    return left == right


def _compare_smaller(left, comparator, right):
    """
    Return True if the left value as integer is smaller than the right integer
    """
    return int(left or 0) < int(right)


def _compare_bigger(left, comparator, right):
    """
    Return True if the left value as integer is bigger than the right integer
    """
    return int(left or 0) > int(right)


def _compare_contains(left, comparator, right):
    """
    Return True if ``left`` has ``right`` as an element.
    Raise a CompareError if ``left`` is not a list.
    :param left: a Python list
    :param right: an arbitrary Python value
    :return: True or False
    """
    if isinstance(left, list):
        return right in left
    else:
        raise CompareError("Left value must be a list, not {!r}".format(type(left)))


def _compare_matches(left, comparator, right):
    """
    Return True if the string in ``left`` completely matches the regular expression given in ``right``.
    Raise a CompareError if ``right`` is not a valid regular expression, or
    if any other matching error occurs.
    :param left: a string
    :param right: a regular expression
    :return: True or False
    """
    try:
        # check for regex modes
        m = re.match(r'^(\(\?[a-zA-Z]+\))(.+)$', right)
        if m and len(m.groups()) == 2:
            regex = m.group(1) + r'^' + m.group(2) + r'$'
        else:
            regex = r"^" + right + r"$"
        return re.match(regex, left) is not None
    except re.error as e:
        raise CompareError("Error during matching: {!r}".format(e))


def _compare_in(left, comparator, right):
    """
    Return True if ``left`` is a member of ``right``, which is a string containing a
    list of values, separated by commas (see ``parse_comma_separated_string``).
    :param left: a string
    :param right: a string of comma-separated values
    :return: True or False
    """
    return left in parse_comma_separated_string(right)


def negate(func):
    """
    Given a comparison function ``func``, build and return a comparison function that negates
    the result of ``func``.
    :param func: a comparison function taking three arguments
    :return: a comparison function taking three arguments
    """
    @wraps(func)
    def negated(left, comparator, right):
        return not func(left, comparator, right)
    return negated


#: This class enumerates all available comparators.
#: In order to add a comparator to this module, add a suitable member to COMPARATORS
#: and suitable entries to COMPARATOR_FUNCTIONS and COMPARATOR_DESCRIPTIONS.
class COMPARATORS(object):
    EQUALS = "equals"
    NOT_EQUALS = "!equals"

    CONTAINS = "contains"
    NOT_CONTAINS = "!contains"

    MATCHES = "matches"
    NOT_MATCHES = "!matches"

    IN = "in"
    NOT_IN = "!in"

    SMALLER = "<"
    BIGGER = ">"


#: This dictionary connects comparators to comparator functions.
#: A comparison function takes three parameters ``left``, ``comparator``, ``right``.
COMPARATOR_FUNCTIONS = {
    COMPARATORS.EQUALS: _compare_equality,
    COMPARATORS.NOT_EQUALS: negate(_compare_equality),

    COMPARATORS.CONTAINS: _compare_contains,
    COMPARATORS.NOT_CONTAINS: negate(_compare_contains),

    COMPARATORS.MATCHES: _compare_matches,
    COMPARATORS.NOT_MATCHES: negate(_compare_matches),

    COMPARATORS.IN: _compare_in,
    COMPARATORS.NOT_IN: negate(_compare_in),

    COMPARATORS.SMALLER: _compare_smaller,
    COMPARATORS.BIGGER: _compare_bigger
}


#: This dictionary connects comparators to their human-readable (and translated) descriptions.
COMPARATOR_DESCRIPTIONS = {
    COMPARATORS.CONTAINS: _("true if the value of the left attribute contains the right value"),
    COMPARATORS.NOT_CONTAINS: _("false if the value of the left attribute contains the right value"),

    COMPARATORS.EQUALS: _("true if the value of the left attribute equals the right value"),
    COMPARATORS.NOT_EQUALS: _("false if the value of the left attribute equals the right value"),

    COMPARATORS.MATCHES: _("true if the value of the left attribute completely matches the given regular expression pattern on the right"),
    COMPARATORS.NOT_MATCHES: _("false if the value of the left attribute completely matches the given regular expression pattern on the right"),

    COMPARATORS.IN: _("true if the value of the left attribute is contained in the comma-separated values on the right"),
    COMPARATORS.NOT_IN: _("false if the value of the left attribute is contained in the comma-separated values on the right"),

    COMPARATORS.SMALLER: _("true if the integer value of the left attribute is smaller than the right integer value"),
    COMPARATORS.BIGGER: _("true if the integer value of the left attribute is bigger than the right integer value")
}


def compare_values(left, comparator, right):
    """
    Compare two values according to ``comparator`` and return either True or False.
    If the comparison is invalid, raise a CompareError with a descriptive message.
    :param left: Left operand of the comparison
    :param comparator: Comparator to use, one of ``COMPARATORS``
    :param right: Right operand of the comparison
    :return: True or False
    """
    if comparator in COMPARATOR_FUNCTIONS:
        return COMPARATOR_FUNCTIONS[comparator](left, comparator, right)
    else:
        raise CompareError("Invalid comparator: {!r}".format(comparator))



"""
This file tests the authentication of admin users and normal user (
selfservice) on the REST API.

implementation is contained in api/auth.py, api/token.py api/audit.py
"""
import datetime
import json

from . import ldap3mock
from .test_api_validate import LDAPDirectory
from .base import MyApiTestCase
from privacyidea.lib.token import (get_tokens, remove_token, enable_token,
                                   assign_token, init_token)
from privacyidea.lib.user import User
from privacyidea.lib.tokenclass import AUTH_DATE_FORMAT
from privacyidea.lib.resolver import save_resolver
from privacyidea.models import Token
from privacyidea.lib.realm import (set_realm, delete_realm, set_default_realm)
from privacyidea.api.lib.postpolicy import DEFAULT_POLICY_TEMPLATE_URL
from privacyidea.lib.policy import (ACTION, SCOPE, set_policy, delete_policy,
                                    LOGINMODE, ACTIONVALUE)


PWFILE = "tests/testdata/passwords"


class APIAuthTestCase(MyApiTestCase):
    """
    This tests some side functionalities of the /auth API.
    """

    def test_00_missing_username(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

    def test_00_missing_password(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "admin"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

    def test_01_get_rights(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            self.at_admin = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "admin", result)

        with self.app.test_request_context('/auth/rights',
                                           method='GET',
                                           data={"username": "testadmin",
                                                 "password": "testpw"},
                                           headers={'Authorization':
                                                        self.at_admin}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("4eyes" in result.get("value"))
            self.assertTrue("hotp" in result.get("value"))
            self.assertTrue(result.get("status"), res.data)

    def test_02_REMOTE_USER(self):
        # Allow remote user
        set_policy(name="remote", scope=SCOPE.WEBUI, action="{0!s}=allowed".format(
                                                            ACTION.REMOTE_USER))

        # Admin remote user
        with self.app.test_request_context('/auth', method='POST',
                                           data={"username": "testadmin"},
                                           environ_base={"REMOTE_USER":
                                                             "testadmin"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("token" in result.get("value"))
            self.assertTrue("username" in result.get("value"))
            self.assertEqual(result.get("value").get("role"), "admin")
            self.assertTrue(result.get("status"), res.data)

        # Check if the /auth request writes the policyname "remote" to the audit entry
        with self.app.test_request_context('/audit/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            auditentry = res.json.get("result").get("value").get("auditdata")[0]
            self.assertTrue("remote" in auditentry.get("policies"))

        self.setUp_user_realms()
        # User "cornelius" from the default realm as normale user
        with self.app.test_request_context('/auth', method='POST',
                                           data={"username": "cornelius"},
                                           environ_base={"REMOTE_USER":
                                                             "cornelius"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("token" in result.get("value"))
            self.assertTrue("username" in result.get("value"))
            self.assertEqual(result.get("value").get("role"), "user")
            self.assertTrue(result.get("status"), res.data)

        # Define the superuser_realm: "adminrealm"
        (added, failed) = set_realm("adminrealm",
                                    [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        # user cornelius is a member of the superuser_realm...
        with self.app.test_request_context('/auth', method='POST',
                                           data={"username":
                                                     "cornelius@adminrealm"},
                                           environ_base={"REMOTE_USER":
                                                     "cornelius@adminrealm"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("token" in result.get("value"))
            self.assertTrue("username" in result.get("value"))
            # ...and will have the role admin
            self.assertEqual(result.get("value").get("role"), "admin")
            self.assertTrue(result.get("status"), res.data)

        delete_policy("remote")

    def test_03_realmadmin_get_user(self):
        # check issue #480
        self.setUp_user_realms()
        self.setUp_user_realm2()
        self.setUp_user_realm3()
        # testadmin is only allowed to view users in realm2
        set_policy(name="realmadmin", scope=SCOPE.ADMIN,
                   action=ACTION.USERLIST, realm=self.realm3, user="testadmin")

        with self.app.test_request_context('/user/',
                                           method='GET',
                                           data={},
                                           headers={'Authorization':
                                                        self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            # In the result list should only be users from reso3.
            for user in result.get("value"):
                self.assertEqual(user.get("resolver"), self.resolvername3)

        delete_policy("realmadmin")


class APIAuthChallengeResponse(MyApiTestCase):

    def setUp(self):
        self.setUp_user_realms()
        # New token for the user "selfservice"
        Token("hotp1", "hotp", otpkey=self.otpkey, userid=1004, resolver=self.resolvername1,
              realm=self.realm1).save()
        # Define HOTP token to be challenge response
        set_policy(name="pol_cr", scope=SCOPE.AUTH, action="{0!s}=hotp".format(ACTION.CHALLENGERESPONSE))
        set_policy(name="webuilog", scope=SCOPE.WEBUI, action="{0!s}=privacyIDEA".format(ACTION.LOGINMODE))
        from privacyidea.lib.token import set_pin
        set_pin("hotp1", "pin")

    def tearDown(self):
        remove_token('hotp1')
        delete_policy('pol_cr')
        delete_policy('webuilog')

    def test_01_challenge_response_at_webui(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice",
                                                 "password": "pin"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            self.assertTrue(data.get("result").get("status"))
            detail = data.get("detail")
            self.assertTrue("enter otp" in detail.get("message"), detail.get("message"))
            transaction_id = detail.get("transaction_id")

        # Now we try to login with the OTP value
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice",
                                                 "password": self.valid_otp_values[0],
                                                 "transaction_id": transaction_id}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            self.assertEqual(data.get("result").get("value").get("username"), "selfservice")

    def test_02_auth_chal_resp_multi_token(self):
        # create another token
        Token("hotp2", "hotp", otpkey=self.otpkey, userid=1004,
              resolver=self.resolvername1, realm=self.realm1).save()
        set_policy(name="pol_otppin", scope=SCOPE.AUTH,
                   action="{0!s}={1!s}".format(ACTION.OTPPIN, ACTIONVALUE.USERSTORE))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice",
                                                 "password": 'test'}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get('detail')
            self.assertTrue('multi_challenge' in result, result)
            self.assertEqual({'hotp1', 'hotp2'}, set([x['serial'] for x in result.get('multi_challenge')]))

        # check if we have both serials in the audit log
        ae = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertTrue({"hotp1", "hotp2"}.issubset(set(ae.get('serial').split(','))), ae)

        delete_policy('pol_otppin')
        remove_token('hotp2')


class APISelfserviceTestCase(MyApiTestCase):

    my_serial = "myToken"
    foreign_serial = "notMyToken"

    def setUp(self):
        """
        For each test we need to initialize the self.at and the self.at_user
        members.
        """
        self.setUp_user_realms()
        Token(self.my_serial, tokentype="hotp", userid="1004",
              resolver="resolver1", realm="realm1").save()
        Token(self.foreign_serial, tokentype="hotp").save()


    def tearDown(self):
        remove_token(self.my_serial)
        remove_token(self.foreign_serial)

    def test_00_authenticate_admin_fail(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "admin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)


    def test_01_authenticate_admin(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            self.at_admin = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "admin", result)

    def test_01_authenticate_admin_from_realm(self):
        # Define an admin realm!
        (added, failed) = set_realm("adminrealm",
                                    [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        # "selfservice" is a user in adminrealm. He should be able to
        # authenticate
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username":
                                                     "selfservice@adminrealm",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            self.at_admin = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "admin", result)

        delete_realm("adminrealm")

    def test_02_user_allowed_to_get_config(self):
        self.authenticate_selfservice_user()
        # The user is allowed to get the system config
        with self.app.test_request_context('/system/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

    def test_02_user_not_allowed(self):
        self.authenticate_selfservice_user()
        # The user is not allowed to write system information
        with self.app.test_request_context('/system/setConfig',
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/system/setDefault',
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/system/documentation',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/system/hsm',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/system/hsm',
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        with self.app.test_request_context('/realm/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # The user is allowed to get his own information
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            value = response.get("result").get("value")
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0].get("username"), "selfservice")

        # If he wants to see other information, he still sees his own
        with self.app.test_request_context('/user/?username=*',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            value = response.get("result").get("value")
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0].get("username"), "selfservice")

        with self.app.test_request_context('/audit/',
                                           method='GET',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

    def test_03_user_enroll_token(self):
        self.authenticate_selfservice_user()
        with self.app.test_request_context('/token/init',
                                           method='POST', data={"genkey": 1},
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))
            serial = response.get("detail").get("serial")
            self.assertTrue("OATH" in serial, serial)

        # Check, who is the owner of the new token!
        tokenobject = get_tokens(serial=serial)[0]
        self.assertEqual(tokenobject.token.first_owner.user_id, "1004")
        self.assertEqual(tokenobject.token.first_owner.resolver, "resolver1")

        # user can delete his own token
        with self.app.test_request_context('/token/{0!s}'.format(serial),
                                           method='DELETE',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))
        # check if there is no token left
        tokenobject_list = get_tokens(serial=serial)
        self.assertTrue(len(tokenobject_list) == 0, len(tokenobject_list))

    def test_04_user_can_not_delete_another_token(self):
        self.authenticate_selfservice_user()
        assign_token(self.foreign_serial, User("cornelius", self.realm1))
        with self.app.test_request_context('/token/{0!s}'.format(self.foreign_serial),
                                           method='DELETE',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])
        # check if the token still exists!
        tokenobject_list = get_tokens(serial=self.foreign_serial)
        self.assertTrue(len(tokenobject_list) == 1, len(tokenobject_list))

    def test_04_user_can_not_disable_another_token(self):
        self.authenticate_selfservice_user()
        assign_token(self.foreign_serial, User("cornelius", self.realm1))
        with self.app.test_request_context('/token/disable/{0!s}'.format(
                                                   self.foreign_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])
        # check if the token still is enabled!
        tokenobject_list = get_tokens(serial=self.foreign_serial)
        self.assertTrue(len(tokenobject_list) == 1, len(tokenobject_list))
        self.assertTrue(tokenobject_list[0].token.active)

    def test_04_user_can_not_lost_another_token(self):
        self.authenticate_selfservice_user()
        assign_token(self.foreign_serial, User("cornelius", self.realm1))
        with self.app.test_request_context('/token/lost/{0!s}'.format(
                                                   self.foreign_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 400, res)


    def test_05_user_can_disable_token(self):
        self.authenticate_selfservice_user()
        # User can not disable a token, that does not belong to him.
        with self.app.test_request_context('/token/disable/{0!s}'.format(
                                                   self.foreign_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])

        tokenobject = get_tokens(serial=self.foreign_serial)[0]
        self.assertTrue(tokenobject.token.active, tokenobject.token.active)

        # User disables his token
        with self.app.test_request_context('/token/disable/{0!s}'.format(self.my_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))

        tokenobject = get_tokens(serial=self.my_serial)[0]
        self.assertFalse(tokenobject.token.active, tokenobject.token.active)

        # User enables his token
        with self.app.test_request_context('/token/enable/{0!s}'.format(self.my_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))

        tokenobject = get_tokens(serial=self.my_serial)[0]
        self.assertTrue(tokenobject.token.active, tokenobject.token.active)

        # User can not enable foreign token
        enable_token(self.foreign_serial, enable=False)
        # Is token disabled?
        tokenobject = get_tokens(serial=self.foreign_serial)[0]
        self.assertFalse(tokenobject.token.active, tokenobject.token.active)
        with self.app.test_request_context('/token/enable/{0!s}'.format(
                                                   self.foreign_serial),
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])

        # token still inactive
        tokenobject = get_tokens(serial=self.foreign_serial)[0]
        self.assertFalse(tokenobject.token.active, tokenobject.token.active)

    def test_06_user_can_assign_token(self):
        self.authenticate_selfservice_user()
        # The foreign token ist not assigned yet, so he can assign it
        with self.app.test_request_context('/token/assign',
                                           data={"serial": self.foreign_serial},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))

        tokenobject = get_tokens(serial=self.foreign_serial)[0]
        self.assertTrue(tokenobject.token.first_owner.user_id == "1004",
                         tokenobject.token.first_owner.user_id)

        # User can unassign token
        with self.app.test_request_context('/token/unassign',
                                           data={"serial": self.foreign_serial},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))

        tokenobject = get_tokens(serial=self.foreign_serial)[0]
        self.assertEqual(tokenobject.token.first_owner, None)


        # User can not unassign token, which does not belong to him
        with self.app.test_request_context('/token/unassign',
                                           data={"serial": self.foreign_serial},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)

    def test_07_user_can_reset_failcount(self):
        self.authenticate_selfservice_user()
        mT = get_tokens(serial=self.my_serial)[0]
        fT = get_tokens(serial=self.foreign_serial)[0]

        fT.token.failcount = 12
        fT.save()

        mT.token.failcount = 12
        mT.save()

        # can not reset foreign token
        with self.app.test_request_context('/token/reset',
                                           data={"serial": self.foreign_serial},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])

        # failcounter still on
        self.assertTrue(fT.token.failcount == 12, fT.token.failcount)

        # can reset own token
        with self.app.test_request_context('/token/reset',
                                           data={"serial": self.my_serial},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))
        # failcounter still on
        self.assertTrue(mT.token.failcount == 0, mT.token.failcount)

    def test_08_user_can_set_pin(self):
        self.authenticate_selfservice_user()
        # can not set pin of foreign token
        with self.app.test_request_context('/token/setpin',
                                           data={"serial":
                                                     self.foreign_serial,
                                                 "otppin": "1234"},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            response = res.json
            self.assertFalse(response["result"]["status"])

        # can set pin for own token
        with self.app.test_request_context('/token/setpin',
                                           data={"serial": self.my_serial,
                                                 "otppin": "1234"},
                                           method='POST',
                                           headers={'Authorization':
                                                        self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            response = res.json
            self.assertTrue(response.get("result").get("value"),
                            response.get("result"))

    def test_09_authz_user_detail(self):
        # Test behavior of ADDUSERINRESPONSE and ADDRESOLVERINRESPONSE policy action for /auth endpoint
        # Normally, no userinfo and realm/resolver are found in the response
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)
            # check that this is a user
            role = result.get("value").get("role")
            self.assertEqual(role, "user", result)
            self.assertEqual(result.get("value").get("realm"), "realm1", content)
            self.assertNotIn("detail", content, content)

        set_policy(name="pol_add_info",
                   scope=SCOPE.AUTHZ,
                   action=[ACTION.ADDUSERINRESPONSE, ACTION.ADDRESOLVERINRESPONSE],
                   realm="realm1")

        # Userinfo + realm/resolver are added to the response
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)
            # check that this is a user
            role = result.get("value").get("role")
            self.assertEqual(role, "user", result)
            self.assertEqual(result.get("value").get("realm"), "realm1", content)
            self.assertEqual(content["detail"]["user-realm"], "realm1", content)
            self.assertEqual(content["detail"]["user-resolver"], "resolver1", content)
            self.assertIn("user", content["detail"], content)
            self.assertEqual(content["detail"]["user"]["userid"], "1004", content)

        # ... but not for internal admins
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)
            self.assertNotIn("detail", content, content)

        delete_policy("pol_add_info")

    def test_10_authz_lastauth(self):
        # Test LASTAUTH policy action for /auth endpoint
        # This only works if we authenticate against privacyIDEA
        set_policy("pol_lastauth",
                   scope=SCOPE.AUTHZ,
                   action={
                       ACTION.LASTAUTH: "10m",
                   })
        set_policy("pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action={
                       ACTION.LOGINMODE: LOGINMODE.PRIVACYIDEA,
                   })
        selfservice_token = init_token({"type": "spass", "pin": "somepin"},
                                       user=User("selfservice", "realm1"))
        # Last authentication was too long ago.
        selfservice_token.add_tokeninfo(ACTION.LASTAUTH, "2016-10-10 10:10:10.000")
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            content = res.json
            result = content.get("result")
            self.assertFalse(result.get("status"), content)
            self.assertIn("long ago", content["detail"]["message"], content)

        selfservice_token.add_tokeninfo(ACTION.LASTAUTH, datetime.datetime.now().strftime(AUTH_DATE_FORMAT))

        # But now it works
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        # Authentication still works for internal admins
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        remove_token(selfservice_token.token.serial)
        delete_policy("pol_lastauth")
        delete_policy("pol_loginmode")

    def test_11_authz_tokentype(self):
        # Check TOKENTYPE policy action for /auth endpoint
        spass_token = init_token({"type": "spass", "pin": "somepin"},
                                       user=User("selfservice", "realm1"))
        set_policy("pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action={
                       ACTION.LOGINMODE: LOGINMODE.PRIVACYIDEA,
                   })
        set_policy("pol_tokentype",
                   scope=SCOPE.AUTHZ,
                   action={
                       ACTION.TOKENTYPE: "hotp",
                   })

        # Cannot authenticate with SPASS token
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            content = res.json
            result = content.get("result")
            self.assertFalse(result.get("status"), content)
            self.assertIn("Tokentype not allowed", result["error"]["message"], content)

        # Allow HOTP+SPASS
        set_policy("pol_tokentype",
                   scope=SCOPE.AUTHZ,
                   action={
                       ACTION.TOKENTYPE: "hotp spass",
                   })
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        # Authentication still works for internal admins
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        remove_token(spass_token.token.serial)
        delete_policy("pol_tokentype")
        delete_policy("pol_loginmode")

    def test_12_authz_tokeninfo(self):
        # Check TOKENINFO policy action for /auth endpoint
        spass_token = init_token({"type": "spass", "pin": "somepin"},
                                 user=User("selfservice", "realm1"))
        set_policy("pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action={
                       ACTION.LOGINMODE: LOGINMODE.PRIVACYIDEA,
                   })
        set_policy("pol_tokeninfo",
                   scope=SCOPE.AUTHZ,
                   action={
                       ACTION.TOKENINFO: "secure/yes/",
                   })

        # Cannot authenticate with token that does not have a "secure" tokeninfo
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            content = res.json
            result = content.get("result")
            self.assertFalse(result.get("status"), content)
            self.assertIn("Tokeninfo field", result["error"]["message"], content)

        # Add the tokeninfo, we can authenticate
        spass_token.add_tokeninfo("secure", "yes")
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        # Authentication still works for internal admins
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        remove_token(spass_token.token.serial)
        delete_policy("pol_tokeninfo")
        delete_policy("pol_loginmode")

    def test_13_authz_tokenserial(self):
        # Check SERIAL policy action for /auth endpoint
        spass_token = init_token({"type": "spass", "pin": "somepin"},
                                 user=User("selfservice", "realm1"))
        set_policy("pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action={
                       ACTION.LOGINMODE: LOGINMODE.PRIVACYIDEA,
                   })
        set_policy("pol_serial",
                   scope=SCOPE.AUTHZ,
                   action={
                       ACTION.SERIAL: "GOOD.*"
                   })

        # Cannot authenticate with token that does not have a "GOOD..." serial
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            content = res.json
            result = content.get("result")
            self.assertFalse(result.get("status"), content)
            self.assertIn("Serial is not allowed for authentication",
                          result["error"]["message"], content)

        # Add a token with a suitable serial
        good_token = init_token({"type": "spass", "pin": "anotherpin", "serial": "GOOD1234"},
                                 user=User("selfservice", "realm1"))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "anotherpin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        # Authentication still works for internal admins
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            content = res.json
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        remove_token(spass_token.token.serial)
        remove_token(good_token.token.serial)
        delete_policy("pol_serial")
        delete_policy("pol_loginmode")

    def test_13_authz_no_detail_on_success(self):
        # Check NODETAILSUCCESS policy action for /auth endpoint
        spass_token = init_token({"type": "spass", "pin": "somepin"},
                                 user=User("selfservice", "realm1"))
        set_policy("pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action={
                       ACTION.LOGINMODE: LOGINMODE.PRIVACYIDEA,
                   })

        # Without the policy, there are details in the response
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            self.assertIn("detail", content, content)
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        # With the policy, there aren't
        set_policy("pol_detail",
                   scope=SCOPE.AUTHZ,
                   action=ACTION.NODETAILSUCCESS)
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "somepin"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            content = res.json
            self.assertNotIn("detail", content, content)
            result = content.get("result")
            self.assertTrue(result.get("status"), content)

        remove_token(spass_token.token.serial)
        delete_policy("pol_detail")
        delete_policy("pol_loginmode")

    def test_31_user_is_not_allowed_for_some_api_calls(self):
        self.authenticate_selfservice_user()
        serial = "serial0001"
        tok = Token(serial)
        tok.save()

        # Can not set things
        with self.app.test_request_context('/token/set',
                                            method="POST",
                                            data={"serial": serial,
                                                  "pin": "test"},
                                            headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # Can not set token realm
        with self.app.test_request_context('/token/realm/{0!s}'.format(serial),
                                            method="POST",
                                            data={"realms": "realm1"},
                                            headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # Can not call get_serial token
        with self.app.test_request_context('/token/getserial/12345',
                                           method="GET",
                                           headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # Can not copy pin
        with self.app.test_request_context('/token/copypin',
                                            method="POST",
                                            headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # Can not copy user
        with self.app.test_request_context('/token/copyuser',
                                            method="POST",
                                            headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # Can not load tokens
        with self.app.test_request_context('/token/load/test.xml',
                                            method="POST",
                                            headers={'Authorization':
                                                         self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

    def test_41_webui_settings(self):
        set_policy(name="webui1", scope=SCOPE.WEBUI, action="{0!s}={1!s}".format(
            ACTION.TOKENPAGESIZE, 20))
        set_policy(name="webui2", scope=SCOPE.WEBUI, action="{0!s}={1!s}".format(
            ACTION.USERPAGESIZE, 20))
        set_policy(name="webui3", scope=SCOPE.WEBUI, action="{0!s}={1!s}".format(
            ACTION.LOGOUTTIME, 200))
        set_policy(name="webui4", scope=SCOPE.WEBUI, action="{0!s}={1!s}".format(
            ACTION.AUDITPAGESIZE, 20))
        set_policy(name="webui5", scope=SCOPE.WEBUI, action=ACTION.DELETION_CONFIRMATION)
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # Test logout time
            self.assertEqual(result.get("value").get("logout_time"), 200)
            self.assertEqual(result.get("value").get("audit_page_size"), 20)
            self.assertEqual(result.get("value").get("token_page_size"), 20)
            self.assertEqual(result.get("value").get("user_page_size"), 20)
            self.assertEqual(result.get("value").get("policy_template_url"),
                             DEFAULT_POLICY_TEMPLATE_URL)
            # check if we have the same values in the data attribute
            result2 = json.loads(res.data.decode('utf8')).get('result')
            self.assertTrue(result2.get("status"), res.data)
            # Test logout time
            self.assertEqual(result2.get("value").get("logout_time"), 200)
            # check if value is True if deletion_confirmation is set
            self.assertEqual(result.get("value").get("deletion_confirmation"), True)

        delete_policy("webui5")

    def test_42_auth_timelimit_maxfail(self):
        self.setUp_user_realm2()
        # check that AUTHMAXFAIL also takes effect for /auth with loginmode=privacyIDEA
        user = User("timelimituser", realm=self.realm2)
        pin = "spass"
        # create a token
        token = init_token({"type": "spass", "pin": pin}, user=user)

        set_policy(name="pol_time1",
                   scope=SCOPE.AUTHZ,
                   action="{0!s}=2/20s".format(ACTION.AUTHMAXFAIL))
        set_policy(name="pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action="{}={}".format(ACTION.LOGINMODE, LOGINMODE.PRIVACYIDEA))
        for _ in range(2):
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "timelimituser@" + self.realm2,
                                                     "password": "wrong"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, 401)

        # We now cannot authenticate even with the correct PIN
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "timelimituser@" + self.realm2,
                                                 "password": pin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            details = res.json.get("detail")
            self.assertEqual(details.get("message"),
                             "Only 2 failed authentications per 0:00:20",
                             details)

        # and even /validate/check does not work
        # (since it counts /auth *and* /validate/check )
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "timelimituser@" + self.realm2,
                                                 "pass": pin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertFalse(result["value"], result)

        delete_policy("pol_time1")
        delete_policy("pol_loginmode")

    def test_43_auth_timelimit_maxsuccess(self):
        self.setUp_user_realm2()
        # check that AUTHMAXSUCCESS also takes effect for /auth with loginmode=privacyIDEA
        user = User("timelimituser", realm=self.realm2)
        pin = "spass"
        # create a token
        token = init_token({"type": "spass", "pin": pin}, user=user)

        set_policy(name="pol_time1",
                   scope=SCOPE.AUTHZ,
                   action="{0!s}=2/20s".format(ACTION.AUTHMAXSUCCESS))
        set_policy(name="pol_loginmode",
                   scope=SCOPE.WEBUI,
                   action="{}={}".format(ACTION.LOGINMODE, LOGINMODE.PRIVACYIDEA))
        for _ in range(2):
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "timelimituser@" + self.realm2,
                                                     "password": pin}):
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, 200)

        # We now cannot authenticate even with the correct PIN
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "timelimituser@" + self.realm2,
                                                 "password": pin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            details = res.json.get("detail")
            self.assertEqual(details.get("message"),
                             "Only 2 successful authentications per 0:00:20",
                             details)

        # ... and not with the wrong PIN
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "timelimituser@" + self.realm2,
                                                 "password": "wrong"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)

        # /validate/check does not work, since the two allowed authentications
        # are already used up for the /auth endpoint
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "timelimituser@" + self.realm2,
                                                 "pass": pin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertFalse(result["value"], result)

        delete_policy("pol_time1")
        delete_policy("pol_loginmode")


class PolicyConditionsTestCase(MyApiTestCase):
    @ldap3mock.activate
    def test_00_set_ldap_realm(self):
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        params = {'LDAPURI': 'ldap://localhost',
                  'LDAPBASE': 'o=test',
                  'BINDDN': 'cn=manager,ou=example,o=test',
                  'BINDPW': 'ldaptest',
                  'LOGINNAMEATTRIBUTE': 'cn',
                  'LDAPSEARCHFILTER': '(cn=*)',
                  'MULTIVALUEATTRIBUTES': '["groups"]',
                  'USERINFO': '{ "username": "cn",'
                                  '"phone" : "telephoneNumber", '
                                  '"mobile" : "mobile"'
                                  ', "email" : "mail", '
                                  '"surname" : "sn", '
                                  '"groups": "memberOf", '
                                  '"givenname" : "givenName" }',
                  'UIDTYPE': 'DN',
                  "resolver": "ldapgroups",
                  "type": "ldapresolver"}

        r = save_resolver(params)
        self.assertTrue(r > 0)

        r = set_realm("ldaprealm", resolvers=[{'name': "ldapgroups"}])
        set_default_realm("ldaprealm")
        self.assertEqual(r, (["ldapgroups"], []))

        # find a user, check the groups
        alice = User("alice", "ldaprealm")
        self.assertEqual(alice.info["groups"], ["cn=admins,o=test", "cn=users,o=test"])

    @ldap3mock.activate
    def test_01_policies_with_userinfo_conditions(self):
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        # create some webui policies with conditions: Depending on their LDAP group membership,
        # users cannot log in at all, are authenticated against privacyIDEA, or against the userstore.

        # disabled policy: by default, login is disabled
        with self.app.test_request_context('/policy/disabled',
                                           json={'action': "{}={}".format(ACTION.LOGINMODE, LOGINMODE.DISABLE),
                                                 'scope': SCOPE.WEBUI,
                                                 'realm': '',
                                                 'priority': 2,
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # userstore policy: for admins, require the userstore password
        with self.app.test_request_context('/policy/userstore',
                                           json={'action': "{}={}".format(ACTION.LOGINMODE, LOGINMODE.USERSTORE),
                                                 'scope': SCOPE.WEBUI,
                                                 'realm': '',
                                                 'priority': 1,
                                                 'conditions': [
                                                     ["userinfo", "groups", "contains", "cn=admins,o=test", True],
                                                 ],
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # privacyidea policy: for helpdesk users, require the token PIN
        with self.app.test_request_context('/policy/privacyidea',
                                           json={'action': "{}={}".format(ACTION.LOGINMODE, LOGINMODE.PRIVACYIDEA),
                                                 'scope': SCOPE.WEBUI,
                                                 'realm': '',
                                                 'priority': 1,
                                                 'conditions': [
                                                     ["userinfo", "groups", "contains", "cn=helpdesk,o=test", True],
                                                 ],
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # enroll 4 SPASS tokens
        users_pins = [("alice", "pin1"),
                      ("bob", "pin2"),
                      ("manager", "pin3"),
                      ("frank", "pin4")]
        for username, pin in users_pins:
            init_token({"type": "spass", "pin": pin}, user=User(username, "ldaprealm"))

        # check our policies:
        # frank is not allowed to log in, because he is neither admin nor helpdesk
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "frank",
                                                 "password": "ldaptest"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = res.json.get("result")
            self.assertFalse(result.get("status"))
            self.assertEqual(result["error"]["message"], "The login for this user is disabled.")

        # alice is allowed to log in with the userstore password, because he is in the admins group
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "alice",
                                                 "password": "alicepw"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # ... but the OTP PIN will not work
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "alice",
                                                 "password": "pin1"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            result = res.json.get("result")
            self.assertFalse(result.get("status"))
            self.assertIn("Wrong credentials", result["error"]["message"])

        # manager can log in with the OTP PIN, because he is in the helpdesk group
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "manager",
                                                 "password": "pin3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # ... but the userstore password will not work
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "manager",
                                                 "password": "ldaptest"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            result = res.json.get("result")
            self.assertFalse(result.get("status"))
            self.assertIn("Wrong credentials", result["error"]["message"])

        # if we now disable the condition on userstore and privacyidea, we get a conflicting policy error
        with self.app.test_request_context('/policy/privacyidea',
                                           json={'scope': SCOPE.WEBUI,
                                                 'action': "{}={}".format(ACTION.LOGINMODE, LOGINMODE.PRIVACYIDEA),
                                                 'realm': '',
                                                 'active': True,
                                                 'conditions': [
                                                     ["userinfo", "groups", "contains", "cn=helpdesk,o=test", False],
                                                 ]},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        with self.app.test_request_context('/policy/userstore',
                                           json={'scope': SCOPE.WEBUI,
                                                 'action': "{}={}".format(ACTION.LOGINMODE, LOGINMODE.USERSTORE),
                                                 'realm': '',
                                                 'active': True,
                                                 'conditions': [
                                                     ["userinfo", "groups", "contains", "cn=admins,o=test", False],
                                                 ]},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # policy error because of conflicting actions
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "manager",
                                                 "password": "pin3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = res.json.get("result")
            self.assertIn("conflicting actions", result["error"]["message"])

        # delete all policies
        delete_policy("disabled")
        delete_policy("userstore")
        delete_policy("privacyidea")

    @ldap3mock.activate
    def test_02_enroll_rights(self):
        # Test a scenario that users are allowed to enroll different tokens according to their LDAP groups
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        # by default, users may only enroll HOTP tokens
        with self.app.test_request_context('/policy/default_enroll',
                                           json={'action': "enrollHOTP",
                                                 'scope': SCOPE.USER,
                                                 'realm': '',
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # but helpdesk users can additionally enroll TOTP tokens
        with self.app.test_request_context('/policy/helpdesk_enroll',
                                           json={'action': "enrollTOTP",
                                                 'scope': SCOPE.USER,
                                                 'realm': '',
                                                 'active': True,
                                                 'conditions': [
                                                     ["userinfo", "groups", "contains", "cn=helpdesk,o=test", True],
                                                 ]},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # check that the endpoint / doesn't throw an error if a user policy with
        # userinfo attribute conditions is defined
        with self.app.test_request_context('/', method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # get an auth token for bob, who is an ordinary user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "bob",
                                                 "password": "bobpwééé"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            bob_token = result['value']['token']
            # check the rights and menus for bob
            self.assertEqual(result['value']['role'], 'user')
            # bob can only enroll HOTP
            self.assertEqual(result['value']['rights'], ['enrollHOTP'])
            self.assertEqual(result['value']['menus'], ['tokens'])

        # get an auth token for manager, who is a helpdesk user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "manager",
                                                 "password": "ldaptest"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            manager_token = result['value']['token']
            # check the rights and menus for manager
            self.assertEqual(result['value']['role'], 'user')
            # manager may enroll HOTP and TOTP
            self.assertEqual(set(result['value']['rights']), {'enrollHOTP', 'enrollTOTP'})
            self.assertEqual(result['value']['menus'], ['tokens'])

        # check the rights of bob
        with self.app.test_request_context('/auth/rights',
                                           method='GET',
                                           headers={'Authorization': bob_token}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            # only HOTP tokens
            self.assertEqual(set(result['value'].keys()), {'hotp'})

        # We set a new policy condition for users (bob), to be allowed to enroll TOTP
        # tokens only in special cases. This condition will not match for certain tokens.
        with self.app.test_request_context('/policy/totp_enroll',
                                           json={'action': "enrollTOTP",
                                                 'scope': SCOPE.USER,
                                                 'realm': '',
                                                 'active': True,
                                                 'conditions': [
                                                     ["userinfo", "username", "equals", "bob", True],
                                                     ["token", "count", "equals", "1", True],
                                                 ]
                                                 },
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # check the rights of bob. The TOTP enroll right will basically be there
        with self.app.test_request_context('/auth/rights',
                                           method='GET',
                                           headers={'Authorization': bob_token}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            # HOTP and TOTP tokens
            self.assertIn('hotp', set(result['value'].keys()))
            self.assertIn('totp', set(result['value'].keys()))

        # check the rights of manager
        with self.app.test_request_context('/auth/rights',
                                           method='GET',
                                           headers={'Authorization': manager_token}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json.get("result")
            # HOTP+TOTP tokens
            self.assertEqual(set(result['value'].keys()), {'hotp', 'totp'})

        # helper function for trying to enroll tokens, returns a tuple (status code, result)
        def _try_enroll(auth_token, token_type):
            with self.app.test_request_context('/token/init',
                                               method='POST',
                                               data={'type': token_type, 'genkey': '1'},
                                               headers={'Authorization': auth_token}):
                res = self.app.full_dispatch_request()
                return res.status_code, res.json.get("result")

        # bob may enroll HOTP ...
        status_code, result = _try_enroll(bob_token, 'hotp')
        self.assertEqual(status_code, 200)
        self.assertEqual(result, {"status": True, "value": True})
        # ... but not TOTP, since the extended condition does not match
        status_code, result = _try_enroll(bob_token, 'totp')
        self.assertEqual(status_code, 403)
        self.assertFalse(result['status'])
        self.assertIn("has conditions on tokens, but a token object is not available", result['error']['message'])
        # ... and certainly not SPASS
        status_code, result = _try_enroll(bob_token, 'spass')
        self.assertEqual(status_code, 403)
        self.assertFalse(result['status'])
        self.assertIn("not allowed to enroll this token type", result['error']['message'])

        # manager may enroll HOTP ...
        status_code, result = _try_enroll(manager_token, 'hotp')
        self.assertEqual(status_code, 200)
        self.assertEqual(result, {"status": True, "value": True})
        # ... and TOTP ...
        status_code, result = _try_enroll(manager_token, 'totp')
        self.assertEqual(status_code, 200)
        self.assertEqual(result, {"status": True, "value": True})
        # ... but not SPASS
        status_code, result = _try_enroll(manager_token, 'spass')
        self.assertEqual(status_code, 403)
        self.assertFalse(result['status'])
        self.assertIn("not allowed to enroll this token type", result['error']['message'])

        delete_policy("default_enroll")
        delete_policy("helpdesk_enroll")



#  2018-08-07   Friedrich Weber <friedrich.weber@netknights.it>
#               Add a shared registry of SQLAlchemy engines to
#               properly implement connection pooling for
#               SQLIdResolvers and SQLAudit modules
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """
This module implements a so-called engine registry which manages
the SQLAlchemy engines used for connections to external SQL databases
by the SQL audit module and SQLIdResolver.

There should only be one shared registry per application which is
used by all threads. This is necessary to properly implement pooling.

This module is tested in tests/test_lib_pooling.py.
"""

import logging
from threading import Lock

from privacyidea.lib.framework import get_app_local_store, get_app_config_value

log = logging.getLogger(__name__)


class BaseEngineRegistry(object):
    """
    Abstract base class for engine registries.
    """
    def get_engine(self, key, creator):
        """
        Return the engine associated with the key ``key``.
        :param key: An arbitrary hashable Python object
        :param creator: A function with no arguments which returns a new SQLAlchemy engine.
                        Called to initially create an engine.
        :return: an SQLAlchemy engine
        """
        raise NotImplementedError()


class NullEngineRegistry(BaseEngineRegistry):
    """
    A registry which creates a new engine for every request.
    Consequently, engines are not shared among threads and
    no pooling is implemented.

    It can be activated by setting ``PI_ENGINE_REGISTRY_CLASS`` to "null".
    """
    def get_engine(self, key, creator):
        return creator()


class SharedEngineRegistry(BaseEngineRegistry):
    """
    A registry which holds a dictionary mapping a key to an SQLAlchemy engine.

    It can be activated by setting ``PI_ENGINE_REGISTRY_CLASS`` to "shared".
    """
    def __init__(self):
        BaseEngineRegistry.__init__(self)
        self._engine_lock = Lock()
        self._engines = {}

    def get_engine(self, key, creator):
        # This method will be called concurrently by multiple threads.
        # Thus, to be sure that we do not create an engine when there
        # is already one associated with the given key, we use a lock.
        with self._engine_lock:
            if key not in self._engines:
                log.info("Creating a new engine and connection pool for key {!s}".format(key))
                self._engines[key] = creator()
            return self._engines[key]


ENGINE_REGISTRY_CLASSES = {
    "null": NullEngineRegistry,
    "shared": SharedEngineRegistry,
}
DEFAULT_REGISTRY_CLASS_NAME = "null"


def get_registry():
    """
    Return the ``EngineRegistry`` object associated with the current application.
    If there is no such object yet, create one and write it to the app-local store.
    This respects the ``PI_ENGINE_REGISTRY_CLASS`` config option.
    :return: an ``EngineRegistry`` object
    """
    # This function will be called concurrently by multiple threads.
    # This is no problem when we already have an engine registry object.
    # However, if there is no registry object yet, two threads may concurrently
    # decide to create a new one. But as ``setdefault`` is atomic, only the
    # first one will be the written to ``app_store['config']``. The latter
    # one will not be referenced and will be garbage-collected at some point.
    app_store = get_app_local_store()
    try:
        return app_store["engine_registry"]
    except KeyError:
        # create a new engine registry of the appropriate class
        registry_class_name = get_app_config_value("PI_ENGINE_REGISTRY_CLASS", DEFAULT_REGISTRY_CLASS_NAME)
        if registry_class_name not in ENGINE_REGISTRY_CLASSES:
            log.warning("Unknown engine registry class: {!r}".format(registry_class_name))
            registry_class_name = DEFAULT_REGISTRY_CLASS_NAME
        registry = ENGINE_REGISTRY_CLASSES[registry_class_name]()
        log.info("Created a new engine registry: {!r}".format(registry))
        return app_store.setdefault("engine_registry", registry)


def get_engine(key, creator):
    """
    Shortcut to get an engine from the application-global engine registry.
    :return: an SQLAlchemy engine
    """
    return get_registry().get_engine(key, creator)


"""
This test file tests the lib.challange methods.

This tests the token functions on an interface level
"""
from .base import MyTestCase
from privacyidea.lib.challenge import get_challenges, extract_answered_challenges, delete_challenges
from privacyidea.lib.policy import (set_policy, delete_policy, SCOPE,
                                    ACTION)
from privacyidea.models import Challenge, db
from privacyidea.lib.token import init_token, check_serial_pass
from privacyidea.lib import _


class ChallengeTestCase(MyTestCase):
    """
    Test the lib.challenge on an interface level
    """

    def test_01_challenge(self):

        set_policy("chalresp", scope=SCOPE.AUTHZ,
                   action="{0!s}=hotp".format(ACTION.CHALLENGERESPONSE))
        token = init_token({"genkey": 1, "serial": "CHAL1", "pin": "pin"})

        r = check_serial_pass(token.token.serial, "pin")
        # The OTP PIN is correct
        self.assertEqual(r[0], False)
        self.assertEqual(r[1].get("message"), _("please enter otp: "))
        transaction_id = r[1].get("transaction_id")
        chals = get_challenges()
        self.assertEqual(len(chals), 1)
        self.assertEqual(chals[0].transaction_id, transaction_id)

        # get challenge for this serial
        chals = get_challenges(serial="CHAL1")
        self.assertEqual(len(chals), 1)
        self.assertEqual(chals[0].transaction_id, transaction_id)

        # get challenge for another seial
        chals = get_challenges(serial="CHAL2")
        self.assertEqual(len(chals), 0)

        delete_policy("chalresp")

    def test_02_extract_answered_challenges(self):
        token = init_token({"genkey": 1, "serial": "CHAL2", "pin": "pin"})
        # no challenges yet
        challenges = get_challenges(serial="CHAL2")
        self.assertEqual(challenges, [])
        self.assertEqual(extract_answered_challenges(challenges), [])
        # we trigger two challenges
        r = check_serial_pass(token.token.serial, "pin")
        self.assertEqual(r[0], False)
        transaction_id1 = r[1].get("transaction_id")
        r = check_serial_pass(token.token.serial, "pin")
        self.assertEqual(r[0], False)
        transaction_id2 = r[1].get("transaction_id")
        # two challenges, but no answered challenges
        challenges = get_challenges(serial="CHAL2")
        self.assertEqual(len(challenges), 2)
        self.assertEqual(extract_answered_challenges(challenges), [])
        # answer one challenge
        Challenge.query.filter_by(transaction_id=transaction_id1).update({"otp_valid": True})
        db.session.commit()
        # two challenges, one answered challenge
        challenges = get_challenges(serial="CHAL2")
        answered = extract_answered_challenges(challenges)
        self.assertEqual(len(challenges), 2)
        self.assertEqual(len(answered), 1)
        self.assertEqual(answered[0].transaction_id, transaction_id1)

    def test_03_delete_challenges(self):
        hotp = init_token({"type": "hotp", "genkey": 1, "pin": "pin"})
        # no challenges yet
        challenges = get_challenges(serial=hotp.get_serial())
        self.assertEqual(challenges, [])

        # trigger two challenges for hotp
        r = check_serial_pass(hotp.get_serial(), "pin")
        self.assertEqual(r[0], False)
        r = check_serial_pass(hotp.get_serial(), "pin")
        self.assertEqual(r[0], False)
        transaction_id = r[1].get("transaction_id")

        # trigger one challenges for totp
        totp = init_token({"type": "totp", "genkey": 1, "pin": "pin"})
        r = check_serial_pass(totp.get_serial(), "pin")
        self.assertEqual(r[0], False)

        # delete challenges by serial
        self.assertEqual(1, len(get_challenges(serial=totp.get_serial())))
        delete_challenges(serial=totp.get_serial())
        self.assertEqual(0, len(get_challenges(serial=totp.get_serial())))

        # delete challenges by transaction_id
        self.assertEqual(1, len(get_challenges(transaction_id=transaction_id)))
        delete_challenges(transaction_id=transaction_id)
        self.assertEqual(0, len(get_challenges(transaction_id=transaction_id)))



"""v3.10: Add container tables (tokencontainer, tokencontainerinfo, tokencontainerowner, tokencontainerrealm,
tokencontainerstates, tokencontainertoken) and add columns container_serial and container_type to the privacy_idea
audit table.

Revision ID: 69e7817b9863
Revises: 2100d1fad908
Create Date: 2024-07-05 10:05:13.987514

"""
from sqlalchemy.exc import OperationalError, ProgrammingError

# revision identifiers, used by Alembic.
revision = '69e7817b9863'
down_revision = '2100d1fad908'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        # ### commands auto generated by Alembic - please adjust! ###
        op.create_table('tokencontainer',
                        sa.Column('id', sa.Integer(), sa.Identity(always=False), nullable=False),
                        sa.Column('type', sa.Unicode(length=100), nullable=False),
                        sa.Column('description', sa.Unicode(length=1024), nullable=True),
                        sa.Column('serial', sa.Unicode(length=40), nullable=False),
                        sa.Column('last_seen', sa.DateTime(), nullable=True),
                        sa.Column('last_updated', sa.DateTime(), nullable=True),
                        sa.PrimaryKeyConstraint('id'),
                        mysql_row_format='DYNAMIC'
                        )
        op.create_index(op.f('ix_tokencontainer_serial'), 'tokencontainer', ['serial'], unique=True)
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainer' already exists.")
        else:
            print("Could not add table 'tokencontainer' to database.")
            print(exx)

    try:
        op.create_table('tokencontainerinfo',
                        sa.Column('id', sa.Integer(), sa.Identity(always=False), nullable=False),
                        sa.Column('key', sa.Unicode(length=255), nullable=False),
                        sa.Column('value', sa.UnicodeText(), nullable=True),
                        sa.Column('type', sa.Unicode(length=100), nullable=True),
                        sa.Column('description', sa.Unicode(length=2000), nullable=True),
                        sa.Column('container_id', sa.Integer(), nullable=True),
                        sa.ForeignKeyConstraint(['container_id'], ['tokencontainer.id'], ),
                        sa.PrimaryKeyConstraint('id'),
                        sa.UniqueConstraint('container_id', 'key', name='container_id_constraint'),
                        mysql_row_format='DYNAMIC'
                        )
        op.create_index(op.f('ix_tokencontainerinfo_container_id'), 'tokencontainerinfo', ['container_id'],
                        unique=False)
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainerinfo' already exists.")
        else:
            print("Could not add table 'tokencontainerinfo' to database.")
            print(exx)

    try:
        op.create_table('tokencontainerowner',
                        sa.Column('id', sa.Integer(), sa.Identity(always=False), nullable=False),
                        sa.Column('container_id', sa.Integer(), nullable=True),
                        sa.Column('resolver', sa.Unicode(length=120), nullable=True),
                        sa.Column('user_id', sa.Unicode(length=320), nullable=True),
                        sa.Column('realm_id', sa.Integer(), nullable=True),
                        sa.ForeignKeyConstraint(['container_id'], ['tokencontainer.id'], ),
                        sa.ForeignKeyConstraint(['realm_id'], ['realm.id'], ),
                        sa.PrimaryKeyConstraint('id'),
                        mysql_row_format='DYNAMIC'
                        )
        op.create_index(op.f('ix_tokencontainerowner_resolver'), 'tokencontainerowner', ['resolver'], unique=False)
        op.create_index(op.f('ix_tokencontainerowner_user_id'), 'tokencontainerowner', ['user_id'], unique=False)
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainerowner' already exists.")
        else:
            print("Could not add table 'tokencontainerowner' to database.")
            print(exx)

    try:
        op.create_table('tokencontainerrealm',
                        sa.Column('container_id', sa.Integer(), nullable=False),
                        sa.Column('realm_id', sa.Integer(), nullable=False),
                        sa.ForeignKeyConstraint(['container_id'], ['tokencontainer.id'], ),
                        sa.ForeignKeyConstraint(['realm_id'], ['realm.id'], ),
                        sa.PrimaryKeyConstraint('container_id', 'realm_id'),
                        mysql_row_format='DYNAMIC'
                        )
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainerrealm' already exists.")
        else:
            print("Could not add table 'tokencontainerrealm' to database.")
            print(exx)

    try:
        op.create_table('tokencontainerstates',
                        sa.Column('id', sa.Integer(), sa.Identity(always=False), nullable=False),
                        sa.Column('container_id', sa.Integer(), nullable=True),
                        sa.Column('state', sa.Unicode(length=100), nullable=False),
                        sa.ForeignKeyConstraint(['container_id'], ['tokencontainer.id'], ),
                        sa.PrimaryKeyConstraint('id'),
                        mysql_row_format='DYNAMIC'
                        )
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainerstates' already exists.")
        else:
            print("Could not add table 'tokencontainerstates' to database.")
            print(exx)

    try:
        op.create_table('tokencontainertoken',
                        sa.Column('token_id', sa.Integer(), nullable=False),
                        sa.Column('container_id', sa.Integer(), nullable=False),
                        sa.ForeignKeyConstraint(['container_id'], ['tokencontainer.id'], ),
                        sa.ForeignKeyConstraint(['token_id'], ['token.id'], ),
                        sa.PrimaryKeyConstraint('token_id', 'container_id'),
                        mysql_row_format='DYNAMIC'
                        )
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Table 'tokencontainertoken' already exists.")
        else:
            print("Could not add table 'tokencontainertoken' to database.")
            print(exx)

    try:
        op.add_column('pidea_audit', sa.Column('container_serial', sa.Unicode(20), nullable=True))
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower() or "duplicate column name" in str(exx.orig).lower():
            print("Columns 'container_serial' already exist.")
        else:
            print("Could not add columns 'container_serial' to table 'pidea_audit'.")
            print(exx)

    try:
        op.add_column('pidea_audit', sa.Column('container_type', sa.Unicode(20), nullable=True))
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower() or "duplicate column name" in str(exx.orig).lower():
            print("Columns 'container_type' already exist.")
        else:
            print("Could not add columns 'container_type' to table 'pidea_audit'.")
            print(exx)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.drop_column('pidea_audit', 'container_type')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such column" in msg or "does not exist" in msg:
            print("Column 'container_type' already removed.")
        else:
            print("Could not remove column 'container_type' from table 'pidea_audit'.")
            print(exx)

    try:
        op.drop_column('pidea_audit', 'container_serial')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such column" in msg or "does not exist" in msg:
            print("Column 'container_serial' already removed.")
        else:
            print("Could not remove column 'container_serial' from table 'pidea_audit'.")
            print(exx)

    try:
        op.drop_table('tokencontainertoken')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainertoken' already removed.")
        else:
            print("Could not remove table 'tokencontainertoken'.")
            print(exx)

    try:
        op.drop_table('tokencontainerstates')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainerstates' already removed.")
        else:
            print("Could not remove table 'tokencontainerstates'.")
            print(exx)

    try:
        op.drop_table('tokencontainerrealm')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainerrealm' already removed.")
        else:
            print("Could not remove table 'tokencontainerrealm'.")
            print(exx)

    try:
        op.drop_index(op.f('ix_tokencontainerowner_user_id'), table_name='tokencontainerowner')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such index" in msg or "does not exist" in msg:
            print("Index 'ix_tokencontainerowner_user_id' already removed.")
        else:
            print("Could not remove index 'ix_tokencontainerowner_user_id' from table 'tokencontainerowner'.")
            print(exx)

    try:
        op.drop_index(op.f('ix_tokencontainerowner_resolver'), table_name='tokencontainerowner')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such index" in msg or "does not exist" in msg:
            print("Index 'ix_tokencontainerowner_resolver' already removed.")
        else:
            print("Could not remove index 'ix_tokencontainerowner_resolver' from table 'tokencontainerowner'.")
            print(exx)

    try:
        op.drop_table('tokencontainerowner')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainerowner' already removed.")
        else:
            print("Could not remove table 'tokencontainerowner'.")
            print(exx)

    try:
        op.drop_index(op.f('ix_tokencontainerinfo_container_id'), table_name='tokencontainerinfo')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such index" in msg or "does not exist" in msg:
            print("Index 'ix_tokencontainerinfo_container_id' already removed.")
        else:
            print("Could not remove index 'ix_tokencontainerinfo_container_id' from table 'tokencontainerinfo'.")
            print(exx)

    try:
        op.drop_table('tokencontainerinfo')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainerinfo' already removed.")
        else:
            print("Could not remove table 'tokencontainerinfo'.")
            print(exx)

    try:
        op.drop_index(op.f('ix_tokencontainer_serial'), table_name='tokencontainer')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such index" in msg or "does not exist" in msg:
            print("Index 'ix_tokencontainer_serial' already removed.")
        else:
            print("Could not remove index 'ix_tokencontainer_serial' from table 'tokencontainer'.")
            print(exx)

    try:
        op.drop_table('tokencontainer')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencontainer' already removed.")
        else:
            print("Could not remove table 'tokencontainer'.")
    # ### end Alembic commands ###



"""Add tokenowner table and move tokenuser data to new table

Revision ID: 48ee74b8a7c8
Revises: cb6d7b7bae63
Create Date: 2019-01-09 16:58:03.968193

"""

# revision identifiers, used by Alembic.
revision = '48ee74b8a7c8'
down_revision = 'cb6d7b7bae63'


from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import Sequence, CreateSequence
from sqlalchemy import orm
from privacyidea.models import TokenRealm, Resolver
import sys

Base = declarative_base()


class Realm(Base):
    __tablename__ = 'realm'
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}
    id = sa.Column(sa.Integer, Sequence("realm_seq"), primary_key=True,
                   nullable=False)
    name = sa.Column(sa.Unicode(255), default=u'',
                     unique=True, nullable=False)
    default = sa.Column(sa.Boolean(), default=False)
    option = sa.Column(sa.Unicode(40), default=u'')


class TokenOwner(Base):
    __tablename__ = 'tokenowner'
    id = sa.Column(sa.Integer(), Sequence("tokenowner_seq"), primary_key=True)
    token_id = sa.Column(sa.Integer(), sa.ForeignKey('token.id'))
    token = orm.relationship('Token', lazy='joined', backref='token_list')
    resolver = sa.Column(sa.Unicode(120), default=u'', index=True)
    user_id = sa.Column(sa.Unicode(320), default=u'', index=True)
    realm_id = sa.Column(sa.Integer(), sa.ForeignKey('realm.id'))
    realm = orm.relationship('Realm', lazy='joined', backref='realm_list')


class Token(Base):
    __tablename__ = 'token'
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}
    id = sa.Column(sa.Integer, Sequence("token_seq"),
                   primary_key=True,
                   nullable=False)
    serial = sa.Column(sa.Unicode(40), default=u'',
                       unique=True,
                       nullable=False,
                       index=True)
    resolver = sa.Column(sa.Unicode(120), default=u'',
                        index=True)
    resolver_type = sa.Column(sa.Unicode(120), default=u'')
    user_id = sa.Column(sa.Unicode(320),
                       default=u'', index=True)


# Check if the SQL dialect uses sequences
# (from https://stackoverflow.com/a/17196812/7036742)
def dialect_supports_sequences():
    migration_context = context.get_context()
    return migration_context.dialect.supports_sequences


def create_seq(seq):
    if dialect_supports_sequences():
        op.execute(CreateSequence(seq))


def upgrade():
    try:
        seq = Sequence('tokenowner_seq')
        try:
            create_seq(seq)
        except Exception as _e:
            pass
        op.create_table('tokenowner',
                        sa.Column('id', sa.Integer(), seq, primary_key=True),
                        sa.Column('token_id', sa.Integer(), nullable=True),
                        sa.Column('resolver', sa.Unicode(length=120), nullable=True),
                        sa.Column('user_id', sa.Unicode(length=320), nullable=True),
                        sa.Column('realm_id', sa.Integer(), nullable=True),
                        sa.ForeignKeyConstraint(['realm_id'], ['realm.id'], ),
                        sa.ForeignKeyConstraint(['token_id'], ['token.id'], ),
                        sa.PrimaryKeyConstraint('id'),
                        mysql_row_format='DYNAMIC'
                        )
        op.create_index(op.f('ix_tokenowner_resolver'), 'tokenowner', ['resolver'], unique=False)
        op.create_index(op.f('ix_tokenowner_user_id'), 'tokenowner', ['user_id'], unique=False)
    except Exception as exx:
        print("Can not create table 'tokenowner'. It probably already exists")
        print (exx)

    try:
        bind = op.get_bind()
        session = orm.Session(bind=bind)
        # For each token, that has an owner, create a tokenowner entry
        for token in session.query(Token).filter(Token.user_id != "", Token.user_id.isnot(None)):
            token_realms = session.query(TokenRealm).filter(TokenRealm.token_id == token.id).all()
            realm_id = None
            if not token_realms:
                sys.stderr.write(u"{serial!s}, {userid!s}, {resolver!s}, "
                                 u"Error while migrating token assignment. "
                                 u"This token has no realm assignments!\n".format(serial=token.serial,
                                                                                  userid=token.user_id,
                                                                                  resolver=token.resolver))
            elif len(token_realms) == 1:
                realm_id = token_realms[0].realm_id
            elif len(token_realms) > 1:
                # The token has more than one realm.
                # In order to figure out the right realm, we first fetch the token's resolver
                resolver = session.query(Resolver).filter_by(name=token.resolver).first()
                if not resolver:
                    sys.stderr.write(u"{serial!s}, {userid!s}, {resolver!s}, "
                                     u"The token is assigned, but the assigned resolver can not "
                                     u"be found!\n".format(serial=token.serial,
                                                           userid=token.user_id,
                                                           resolver=token.resolver))
                else:
                    # Then, fetch the list of ``Realm`` objects in which the token resolver is contained.
                    resolver_realms = [r.realm for r in resolver.realm_list]
                    if not resolver_realms:
                        sys.stderr.write(u"{serial!s}, {userid!s}, {resolver!s}, "
                                         u"The token is assigned, but the assigned resolver is not "
                                         u"contained in any realm!\n".format(serial=token.serial,
                                                                             userid=token.user_id,
                                                                             resolver=token.resolver))
                    elif len(resolver_realms) == 1:
                        # The resolver is only in one realm, so this is the new realm of the token!
                        realm_id = resolver_realms[0].id
                    elif len(resolver_realms) > 1:
                        # The resolver is contained in more than one realm, we have to apply more logic
                        # between the realms in which the resolver is contained and the realms,
                        # to which the token is assigend.
                        # More specifically, we find all realms which are both a token realm and
                        # a realm of the token resolver.
                        # If there is exactly one such realm, we have found our token owner realm.
                        # If there is more than one such realm, we cannot uniquely identify a token owner realm.
                        # If there is no such realm, we have an inconsistent database.
                        found_realm_ids = []
                        found_realm_names = []
                        for token_realm in token_realms:
                            if token_realm.realm in resolver_realms:
                                # The token realm, that also fits the resolver_realm is used as owner realm
                                found_realm_ids.append(token_realm.realm.id)
                                found_realm_names.append(token_realm.realm.name)
                        if len(found_realm_ids) > 1:
                            sys.stderr.write(u"{serial!s}, {userid!s}, {resolver!s}, Can not assign token. "
                                             u"Your realm configuration for the token is not distinct! "
                                             u"The tokenowner could be in multiple realms! "                                  
                                             u"The token is assigned to the following realms and the resolver is also "
                                             u"contained in these realm IDs: {realms!s}.\n".format(serial=token.serial,
                                                                                                   userid=token.user_id,
                                                                                                   resolver=token.resolver,
                                                                                                   realms=found_realm_names))
                        elif len(found_realm_ids) == 1:
                            realm_id = found_realm_ids[0]
                        else:
                            sys.stderr.write(u"{serial!s}, {userid!s}, {resolver!s}, "
                                             u"Can not assign token. The resolver is not contained in any "
                                             u"realms, to which the token is assigned!\n".format(serial=token.serial,
                                                                                                 userid=token.user_id,
                                                                                                 resolver=token.resolver))
            # If we could not figure out a tokenowner realm, we skip the token assignment.
            if realm_id is not None:
                to = TokenOwner(token_id=token.id, user_id=token.user_id,
                                resolver=token.resolver, realm_id=realm_id)
                session.add(to)
        session.commit()

        # Now we drop the columns
        op.drop_column('token', 'user_id')
        op.drop_column('token', 'resolver')
        op.drop_column('token', 'resolver_type')

    except Exception as exx:
        session.rollback()
        print("Failed to migrate token assignment data!")
        print (exx)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_tokenowner_user_id'), table_name='tokenowner')
    op.drop_index(op.f('ix_tokenowner_resolver'), table_name='tokenowner')
    op.drop_table('tokenowner')
    # ### end Alembic commands ###



"""
This file tests the web UI for creating certificate requests

implementation is contained webui/certificate.py
"""
import json
from .base import MyTestCase
import os
from privacyidea.lib.caconnectors.localca import LocalCAConnector


REQUESTKEY = """MIICQDCCASgwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDPPNQzDhLE5trNtlahHCa7JhyKSnOoWIq2HrOLmdIdB8KSiXnAadjI3yhhmB6z/q4rvXip925H3KOgFoyFAsFOkv1ybvDIAymbuABuwIOVyDQgNpyz5eTmdnOjGq1AluBTVADsdnmaxg1+tr0p7IPzy4mky2wAugzFeA//abiU9ARwQz/Ynten+13OdY7a58QHsZ3eb4hLtk2az/m8+p/NMm32OgsNI0J47JdCvw5NYFbh0wLyGcuEV5DlcKGigzWG4tqGn/mKxHmzUijay7s0ytPakUrPjXaismub+Zb9CSraESNN8MvWsrEOEmyaGWWYh8rk7iTORKyQj50bxSqdAgMBAAEWADANBgkqhkiG9w0BAQQFAAOCAQEAjqR8Cv+UZeGXP9v00/T4ClH2wCtQea9oLklllElsU+x9UNjrPITpZGwiKdCtrPSDy+QeqzecSXi23LL05s6RKATnQt31EPRLLPuHRwkpbHD+n/XJtqv5Byge/KJX+Xt8xb+cLKfGJmQibnV/vu83TL8on91pUB4BXuaSu3UXJOFnrG0E2h4rpGE8FrK3JrIruQe2FAcal/KRGzsgHp/vq90OibH0ZJQE3kNg+JkOlzBTTn73+Q39y/E6CW7fD8iFtNRF0xhZYJ/AgflLMQeQUeKRb0Qaillz/DnWQFuVqLoCdahvv6jt58nXmqHv6oMfRg0R2qF2jMFfGtI15Hixzw=="""
CAKEY = "cakey.pem"
CACERT = "cacert.pem"
OPENSSLCNF = "openssl.cnf"
WORKINGDIR = "tests/testdata/ca"


class WebUICertificateTestCase(MyTestCase):

    my_serial = "myToken"
    foreign_serial = "notMyToken"

    def setUp(self):
        """
        For each test we need to initialize the users
        """
        self.setUp_user_realms()

    def test_01_cert_request(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            authtoken = result.get("value")

        # Check the form
        with self.app.test_request_context('/certificate',
                                           method='POST',
                                           data={"authtoken": authtoken.get("token")}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual(res.mimetype, 'text/html', res)
            # Check the form
            self.assertTrue(b"privacyIDEA Certificate Request" in res.data)
            self.assertTrue(b"Key strength" in res.data)
            self.assertTrue(b'input type="hidden" name="authtoken"' in res.data)

        # Check that missing authentication will result in an error
        with self.app.test_request_context('/certificate',
                                           method='POST',
                                           data={}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # GET-Request will fail, Method not allowed
        with self.app.test_request_context('/certificate',
                                           method='GET',
                                           data={}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 405, res)

    def test_02_cert_enrolled(self):
        # Setup the CA
        self.authenticate()
        cwd = os.getcwd()
        with self.app.test_request_context('/caconnector/localCA',
                                           data={'type': 'local',
                                                 'cakey': CAKEY,
                                                 'cacert': CACERT,
                                                 'openssl.cnf': OPENSSLCNF,
                                                 "WorkingDir": cwd + "/" + WORKINGDIR},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json['result']
            self.assertTrue(result["status"], result)
            self.assertTrue(result["value"] == 1, result)

        # Get the users authtoken
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json['result']
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            authtoken = result.get("value")

        # Check the form
        with self.app.test_request_context('/certificate/enroll',
                                           method='POST',
                                           data={"authtoken": authtoken.get("token"),
                                                 "requestkey": REQUESTKEY,
                                                 "ca": "localCA"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual(res.mimetype, 'text/html', res)
            # Check the form
            self.assertTrue(b"The certificate with token serial" in res.data)
            self.assertTrue(b"Certificate to Browser" in res.data)
            self.assertTrue(b'data:application/x-x509-user-cert;base64' in res.data)

        # Check that missing authentication will result in an error
        with self.app.test_request_context('/certificate/enroll',
                                           method='POST',
                                           data={}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # GET-Request will fail, Method not allowed
        with self.app.test_request_context('/certificate/enroll',
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 405, res)



"""
This test file tests the applications definitions standalone
lib/applications/*
"""
from privacyidea.lib.error import ParameterError
from .base import MyTestCase
from privacyidea.lib.applications import MachineApplicationBase
from privacyidea.lib.applications.ssh import (MachineApplication as
                                              SSHApplication)
from privacyidea.lib.applications.luks import (MachineApplication as
                                               LUKSApplication)
from privacyidea.lib.applications.offline import (MachineApplication as
                                                  OfflineApplication,
                                                  REFILLTOKEN_LENGTH)
from privacyidea.lib.applications import (get_auth_item,
                                          is_application_allow_bulk_call,
                                          get_application_types)
from privacyidea.lib.token import init_token, get_tokens
from privacyidea.lib.user import User
import passlib.hash
import mock


SSHKEY = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQDO1rx377" \
         "cmSSs/89j/0u5aEiXa7bYArHn7zFNCBaVnDUiK9JDNkpWB" \
         "j2ucbmOpDKWzH0Vl3in21E8BaRlq9BobASG0qlEqlnwrYwlH" \
         "+vcYp6td4QBoh3sOelzhyrJFug9dnfe8o70r3IL4HIbdQOdh1" \
         "b8Ogi7aL01V/eVE9RgGfTNHUzuYRMUL3si4dtqbCsSjFZ6dN1mm" \
         "Vhos9cSphPr7pQEbq8xW0uxzOGrFDY9g1NSOleA8bOjsCT9k+3X" \
         "4R700iVGvpzWkKopcWrzXJDIa3yxylAMOM0c3uO9U3NLfRsucvc" \
         "Q5Cs8S6ctM308cua3t5WaBOsr3RyoXs+cHIPIkXnJHg03HsnWON" \
         "aGxl8VPymC9s3P0zVwm2jMFxJD9WbCqep7Dwc5unxLOSKidKrnNflQ" \
         "iMyiIv+5dY5lhc0YTJdktC2Scse64ac2E7ldjG3bJuKSIWAz8Sd" \
         "1km4ZJWWIx8NlpC9AfbHcgMyFUDniV1EtFIaSQLPspIkthzIMq" \
         "PTpKblzdRZP37mPu/FpwfYG4S+F34dCmJ4BipslsVcqgCFJQHo" \
         "AYAJc4NDq5IRDQqXH2KybHpSLATnbSY7zjVD+evJeU994yTa" \
         "XTFi5hBmd0aWTC+ph79mmEtu3dokA2YbLa7uWkAIXvX/HHauGLM" \
         "TyCOpYi1BxN47c/kccxyNgjPw== user@example.com"
OTPKEY = "3132333435363738393031323334353637383930"


class SSHApplicationTestCase(MyTestCase):

    def test_01_get_options(self):
        # Can run as class
        tokenoptions = SSHApplication.get_options()
        self.assertIn("sshkey", tokenoptions)
        sshkeyoptions = tokenoptions.get("sshkey")
        self.assertIn("user", sshkeyoptions)
        self.assertIn("service_id", sshkeyoptions)
        self.assertEqual("str", sshkeyoptions.get("user").get("type"))
        self.assertEqual("str", sshkeyoptions.get("service_id").get("type"))

    def test_02_get_auth_item(self):
        serial = "ssh1"
        # create realm
        self.setUp_user_realms()
        user = User("cornelius", realm=self.realm1)
        # create ssh token
        init_token({"serial": serial, "type": "sshkey", "sshkey": SSHKEY},
                   user=user)

        auth_item = SSHApplication.get_authentication_item("sshkey", serial)
        self.assertEqual(auth_item.get("sshkey"), SSHKEY)
        self.assertEqual(auth_item.get("username"), "cornelius")

        # Get with no matching user
        with mock.patch("logging.Logger.debug") as mock_log:
            auth_item = SSHApplication.get_authentication_item("sshkey", serial, filter_param={"user": "Idefix"})
            self.assertFalse(auth_item)
            mock_log.assert_called_with('The requested user Idefix does '
                                        'not match the user option (Idefix) of the SSH application.')

    def test_03_get_auth_item_unsupported(self):
        # unsupported token type
        auth_item = SSHApplication.get_authentication_item("unsupported", "s")
        self.assertEqual(auth_item, {})


class LUKSApplicationTestCase(MyTestCase):

    def test_01_get_options(self):
        # Can run as class
        tokenoptions = LUKSApplication.get_options()
        self.assertIn("totp", tokenoptions)
        totpoptions = tokenoptions.get("totp")
        self.assertIn("slot", totpoptions)
        self.assertIn("partition", totpoptions)
        self.assertEqual("int", totpoptions.get("slot").get("type"))
        self.assertEqual("str", totpoptions.get("partition").get("type"))

    def test_02_get_auth_item(self):
        serial = "UBOM12345"
        # create realm
        self.setUp_user_realms()
        user = User("cornelius", realm=self.realm1)
        # create ssh token
        init_token({"serial": serial, "type": "totp", "otpkey": OTPKEY},
                   user=user)

        auth_item = LUKSApplication.get_authentication_item("totp", serial)
        self.assertEqual(len(auth_item.get("challenge")), 64, auth_item)
        self.assertIsInstance(auth_item.get('challenge'), str, auth_item)
        self.assertEqual(len(auth_item.get("response")), 40, auth_item)

        auth_item = LUKSApplication.get_authentication_item("totp", serial,
                                                            challenge="123456")
        self.assertEqual(auth_item.get("challenge"), "123456")
        self.assertEqual(auth_item.get("response"),
                         "76d624a5fdf8d84f3d19e781f0313e48c1e69165")

    def test_03_get_auth_item_unsupported(self):
        # unsupported token type
        auth_item = LUKSApplication.get_authentication_item("unsupported", "s")
        self.assertEqual(auth_item, {})


class OfflineApplicationTestCase(MyTestCase):

    def test_01_get_options(self):
        # Can run as class
        tokenoptions = OfflineApplication.get_options()
        self.assertIn("hotp", tokenoptions)
        self.assertIn("webauthn", tokenoptions)
        hotpoptions = tokenoptions.get("hotp")
        self.assertIn("count", hotpoptions)
        self.assertIn("rounds", hotpoptions)
        self.assertEqual("str", hotpoptions.get("count").get("type"))
        self.assertEqual("str", hotpoptions.get("rounds").get("type"))

    def test_02_get_auth_item(self):
        serial = "OATH1"
        # create realm
        self.setUp_user_realms()
        user = User("cornelius", realm=self.realm1)
        # create ssh token
        init_token({"serial": serial, "type": "hotp", "otpkey": OTPKEY},
                   user=user)
        # authenticate online initially
        tok = get_tokens(serial=serial)[0]
        res = tok.check_otp("359152") # count = 2
        self.assertEqual(res, 2)
        # check intermediate counter value
        self.assertEqual(tok.token.count, 3)

        auth_item = OfflineApplication.get_authentication_item("hotp", serial)
        refilltoken = auth_item.get("refilltoken")
        self.assertEqual(len(refilltoken), REFILLTOKEN_LENGTH * 2)
        self.assertTrue(passlib.hash.\
                        pbkdf2_sha512.verify("969429", # count = 3
                                             auth_item.get("response").get(3)))
        self.assertTrue(passlib.hash.\
                        pbkdf2_sha512.verify("399871", # count = 8
                                             auth_item.get("response").get(8)))
        # The token now contains the refill token information:
        self.assertEqual(refilltoken, tok.get_tokeninfo("refilltoken"))

        # After calling auth_item the token counter should be increased
        # 3, because we used the otp value with count = 2 initially
        # 100, because we obtained 100 offline OTPs
        self.assertEqual(tok.token.count, 3 + 100)
        # Assert that we cannot authenticate with the last offline OTP we got
        self.assertEqual(len(auth_item.get("response")), 100)
        self.assertTrue(passlib.hash.\
                        pbkdf2_sha512.verify("629694", # count = 102
                                             auth_item.get("response").get(102)))
        res = tok.check_otp("629694") # count = 102
        self.assertEqual(res, -1)
        res = tok.check_otp("378717")  # count = 103
        self.assertEqual(res, 103)
        # check illegal API usage
        self.assertRaises(ParameterError,
                          OfflineApplication.get_offline_otps, tok, 'foo', -1)
        self.assertEqual(OfflineApplication.get_offline_otps(tok, 'foo', 0), {})

    def test_03_get_auth_item_unsupported(self):
        # unsupported token type
        auth_item = OfflineApplication.get_authentication_item("unsupported",
                                                               "s")
        self.assertEqual(auth_item, {})


class BaseApplicationTestCase(MyTestCase):

    def test_01_create_base_application(self):
        base_app = MachineApplicationBase()
        self.assertEqual(base_app.get_name(), "base")
        self.assertEqual(base_app.get_authentication_item("hotp", "serial"),
                         "nothing")
        tokens = base_app.get_options()
        options = tokens.get("tokentype")
        self.assertEqual(options["optionA"].get("type"), "bool")
        self.assertTrue(options["optionA"].get("required"))

    def test_02_get_auth_item(self):
        auth_item = get_auth_item("base", "hotp", "serial")
        self.assertEqual(auth_item, "nothing")

    def test_03_allow_bulk_call(self):
        bulk = is_application_allow_bulk_call(
            "privacyidea.lib.applications.base")
        self.assertFalse(bulk)

    def test_04_get_application_types(self):
        apps = get_application_types()
        self.assertTrue("luks" in apps)
        self.assertTrue("ssh" in apps)
        sshoptions = apps["ssh"]["options"]
        offlineoptions = apps["offline"]["options"]
        luksoptions = apps["luks"]["options"]
        self.assertIn("service_id", sshoptions["sshkey"])
        self.assertIn("user", sshoptions["sshkey"])
        self.assertIn("slot", luksoptions["totp"])
        self.assertIn("partition", luksoptions["totp"])
        self.assertIn("count", offlineoptions["hotp"])
        self.assertIn("rounds", offlineoptions["hotp"])
        self.assertEqual({}, offlineoptions["webauthn"])





# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import base64
import os
import pathlib
import sys
import click
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from flask.cli import AppGroup
from flask import current_app
from flask_migrate import stamp as fm_stamp
import gnupg
from importlib import metadata

from privacyidea.models import db
from privacyidea.lib.security.default import DefaultSecurityModule

setup_cli = AppGroup("setup", short_help="privacyIDEA server setup",
                     help="Commands to setup the privacyIDEA server for production")


@setup_cli.command("encrypt_enckey", short_help="Additionally encrypt the encryption key")
@click.argument("encfile", type=click.File("rb"))
@click.option("-o", "--outfile", type=click.File("w"), default=sys.stdout,
              help="The file to which the encrypted encryption key will be "
                   "written to (default: stdout)")
@click.password_option(help="The password to encrypt the encryption key. "
                            "If this option is not given the user will be "
                            "prompted for the password.")
def encrypt_enckey(encfile, outfile, password):
    """
    You will be asked for a password and the given encryption key in the specified
    file will be encrypted with an AES key derived from your password.

    The encryption key in the file is a 96 bit binary key.

    The password based encrypted encryption key is a hex combination of an IV
    and the encrypted data.

    The result can be piped to a new enckey file.
    """
    # TODO we just print out a string here and assume, the user pipes it into a file.
    #      Maybe we should write the file here so we know what is in there
    # TODO: get the name/path of the encfile from the pi.cfg
    enckey = encfile.read()
    res = DefaultSecurityModule.password_encrypt(enckey, password)
    outfile.write(res)
    outfile.write('\n')


@setup_cli.command("create_enckey")
@click.option("-e", "--enckey_b64", type=str,
              help="base64 encoded plain text key")
@click.pass_context
def create_enckey(ctx, enckey_b64):
    """
    Create a key for encrypting the sensitive database entries.

    If the key of the given configuration does not yet exist, it will be created.
    Optionally a base64 encoded key can be passed with the "--enckey_b64" option.

    Warning: passing the encryption key as a parameter is considered harmful!
    """
    enc_file = pathlib.Path(current_app.config.get("PI_ENCFILE"))
    if enc_file.is_file():
        click.secho(f"The file \n\t{enc_file}\nalready exist. We do not overwrite it!",
                    fg="yellow")
        ctx.exit(1)
    with open(enc_file, "wb") as f:
        if enckey_b64 is None:
            f.write(DefaultSecurityModule.random(96))
        else:
            click.secho("Warning: Passing enckey via cli input is considered harmful.",
                        fg="yellow")
            bin_enckey = base64.b64decode(enckey_b64)
            if len(bin_enckey) != 96:
                click.secho("Error: enckey must be 96 bytes length", fg="red")
                ctx.exit(1)
            f.write(bin_enckey)
    click.secho(f"Encryption key written to {enc_file}", fg="green")
    enc_file.chmod(0o400)
    click.secho(f"The file permission of {enc_file} was set to 400!", fg="yellow")
    click.secho("Please ensure, that it is owned by the right user.", fg="yellow")


@setup_cli.command("create_pgp_keys")
@click.option("-f", "--force", is_flag=True,
              help="Overwrite existing PGP keys")
@click.option("-k", "--keysize", type=int, default=2048, show_default=True,
              help="Size of the generated PGP keys (in bits)")
@click.pass_context
def create_pgp_keys(ctx, keysize, force):
    """
    Generate PGP keys to allow encrypted token import.
    """
    # TODO: change owner and permission of gpg directory
    gpg_home = pathlib.Path(current_app.config.get("PI_GNUPG_HOME", "/etc/privacyidea/gpg"))
    if not gpg_home.exists():
        try:
            gpg_home.mkdir(parents=True)
        except IOError as e:
            click.secho(f"Could not create PGP directory {gpg_home}: {e}", fg="red")
            ctx.exit(1)
    gpg = gnupg.GPG(gnupghome=gpg_home)
    keys = gpg.list_keys(True)
    if len(keys) and not force:
        click.secho("There are already private keys. If you want to generate a "
                    "new private key, use the parameter --force.", fg="yellow")
        click.echo(f"uids: {keys[0]['uids']}\t fingerprint: {keys[0]['fingerprint']}")
        ctx.exit(1)
    else:
        click.secho("Overwriting existing PGP keys!", fg="yellow")
    input_data = gpg.gen_key_input(key_type="RSA", key_length=keysize,
                                   name_real="privacyIDEA Server",
                                   name_comment="Import")
    inputs = input_data.split("\n")
    if inputs[-2] == "%commit":
        del (inputs[-1])
        del (inputs[-1])
        inputs.append("%no-protection")
        inputs.append("%commit")
        inputs.append("")
        input_data = "\n".join(inputs)
    gpg.gen_key(input_data)


@setup_cli.command("create_audit_keys")
@click.option("-k", "--keysize", type=int, default=2048, show_default=True,
              help="Create keys with the given size in bits")
@click.pass_context
def create_audit_keys(ctx, keysize):
    """
    Create the RSA signing keys for the audit log.

    You may specify a different key size.
    The default key size is 2048 bit.
    """
    priv_key = pathlib.Path(current_app.config.get("PI_AUDIT_KEY_PRIVATE"))
    if priv_key.is_file():
        click.secho(f"The file \n\t{priv_key}\nalready exist. We do not overwrite it!",
                    fg="yellow")
        ctx.exit(1)
    new_key = rsa.generate_private_key(public_exponent=65537,
                                       key_size=keysize,
                                       backend=default_backend())
    priv_pem = new_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption())
    with open(priv_key, "wb") as f:
        f.write(priv_pem)

    pub_key = pathlib.Path(current_app.config.get("PI_AUDIT_KEY_PUBLIC"))
    public_key = new_key.public_key()
    pub_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo)
    with open(pub_key, "wb") as f:
        f.write(pub_pem)

    click.secho(f"Signing keys written to {priv_key} and {pub_key}", fg="green")
    priv_key.chmod(0o400)
    click.secho(f"The file permission of {priv_key} was set to 400!", fg="yellow")
    click.secho("Please ensure, that it is owned by the right user.", fg="yellow")


@setup_cli.command("create_tables")
@click.option("-n", "--no-stamp", is_flag=True, default=False,
              show_default=True, help='Do not stamp the database to a revision.')
@click.option("-s", "--stamp", is_flag=True, default=False, hidden=True)
def create_tables(no_stamp, stamp):
    """
    Initially create the tables in the database. The database must exist
    (an SQLite database will be created).
    """
    click.echo(f"Using connect string {db}")
    if stamp:
        click.secho("The parameter \"--stamp\" is deprecated, the database "
                    "will be stamped by default. Use parameter \"--no-stamp\" "
                    "to prevent this.", fg="yellow")
    db.create_all()
    if not no_stamp:
        # get the path to the migration directory from the distribution
        p = [x.locate() for x in metadata.files('privacyidea') if
             'migrations/env.py' in str(x)]
        migration_dir = os.path.dirname(os.path.abspath(p[0]))
        fm_stamp(directory=migration_dir)
    db.session.commit()


@setup_cli.command("drop_tables")
@click.option("-d", "--dropit", type=str,
              help="If You are sure to drop the tables, pass the parameter \"yes\"")
def drop_tables(dropit):
    """
    This drops all the privacyIDEA database tables.
    Use with caution! All data will be lost!

    For safety reason you need to pass "--dropit==yes",
    otherwise the command will not drop anything.
    """
    if dropit == "yes":
        click.echo("Dropping all database tables!")
        db.drop_all()
        table_name = "alembic_version"
        db.reflect()
        table = db.metadata.tables.get(table_name, None)
        if table is not None:
            db.metadata.drop_all(bind=db.engine,
                                 tables=[table],
                                 checkfirst=True)
    else:
        click.echo("Not dropping anything!")
